
// Genode includes
#include <base/heap.h>
#include <base/quota_guard.h>

// gapfruit includes
#include <tee/packet_allocator.h>


namespace Tee
{
	using namespace Genode;

	using Genode::off_t;

	class Session_env;

	inline size_t parse_ram_quota(const char* args) {
		return Genode::Arg_string::find_arg(args, "ram_quota").ulong_value(0);
	}

	inline size_t parse_cap_quota(const char* args) {
		return Genode::Arg_string::find_arg(args, "cap_quota").ulong_value(0);
	}

	inline size_t parse_tx_buffer_size(const char* args) {
		return Genode::Arg_string::find_arg(args, "tx_buf_size").ulong_value(0);
	}

	inline size_t parse_rx_buffer_size(const char* args) {
		return Genode::Arg_string::find_arg(args, "rx_buf_size").ulong_value(0);
	}

}


class Tee::Session_env : public Ram_allocator,
                         public Region_map
{
	private:

		Env                    &_env;
		Ram_quota_guard         _ram_guard;
		Cap_quota_guard         _cap_guard;

		template <typename FUNC>
		void _consume(size_t  ram_amount,
		              size_t  cap_amount,
		              FUNC && functor)
		{
			size_t ram_consumed { _env.pd().used_ram().value };
			size_t cap_consumed { _env.pd().used_caps().value };

			{
				Ram_quota_guard::Reservation ram_reserve { _ram_guard, Ram_quota { ram_amount } };
				Cap_quota_guard::Reservation cap_reserve { _cap_guard, Cap_quota { cap_amount } };

				functor();

				ram_reserve.acknowledge();
				cap_reserve.acknowledge();
			}

			ram_consumed = _env.pd().used_ram().value  - ram_consumed;
			cap_consumed = _env.pd().used_caps().value - cap_consumed;

			if (ram_consumed < ram_amount) {
				error("Session_env: less RAM quota consumed than expected");
			}

			if (cap_consumed < cap_amount) {
				error("Session_env: less CAP quota consumed than expected");
			}

			_ram_guard.replenish( Ram_quota { ram_consumed } );
			_cap_guard.replenish( Cap_quota { cap_consumed } );
		}

		template <typename FUNC>
		void _replenish(size_t  accounted_ram,
		                size_t  accounted_cap,
		                FUNC && functor)
		{
			size_t ram_replenish { _env.pd().used_ram().value };
			size_t cap_replenish { _env.pd().used_caps().value };

			functor();

			ram_replenish = ram_replenish - _env.pd().used_ram().value;
			cap_replenish = cap_replenish - _env.pd().used_caps().value;

			if (ram_replenish < accounted_ram) {
				error("Session_env: less RAM quota replenished than expected");
			}

			if (cap_replenish < accounted_cap) {
				error("Session_env: less CAP quota replenished than expected");
			}

			_ram_guard.replenish( Ram_quota { accounted_ram } );
			_cap_guard.replenish( Cap_quota { accounted_cap } );
		}

	public:

		Session_env(Env &env, Ram_quota const &ram_quota, Cap_quota const &cap_quota) :
			_env       { env },
			_ram_guard { ram_quota },
			_cap_guard { cap_quota }
		{ }

		Entrypoint &ep() { return _env.ep(); }

		/*******************
		 ** Ram_allocator **
		 *******************/
		Ram_dataspace_capability alloc(size_t size, Cache_attribute cached) override
		{
			enum { DS_SIZE_GRANULARITY_LOG2 = 12 };

			const size_t             ds_size { align_addr(size, DS_SIZE_GRANULARITY_LOG2) };
			Ram_dataspace_capability ds_cap;

			_consume(ds_size, 1, [&] () { ds_cap = _env.pd().alloc(ds_size, cached); });

			return ds_cap;
		}

		void free(Ram_dataspace_capability ds) override
		{
			_replenish(_env.pd().dataspace_size(ds), 1, [&] () {
				_env.pd().free(ds);
			});
		}

		size_t dataspace_size(Ram_dataspace_capability ds) const override { return _env.pd().dataspace_size(ds); }

		/****************
		 ** Region_map **
		 ****************/
		Local_addr attach(Dataspace_capability ds,
		                  size_t               size = 0,
		                  off_t                offset = 0,
		                  bool                 use_local_addr = false,
		                  Local_addr           local_addr = nullptr,
		                  bool                 executable = false,
		                  bool                 writeable = true) override
		{
			void *ptr;
			_consume(0, 0, [&] () {
				ptr = _env.rm().attach(ds, size, offset, use_local_addr,
				                       local_addr, executable, writeable);
			});
			return ptr;
		}

		void detach(Local_addr local_addr) override
		{
			_replenish(0, 0, [&] () { _env.rm().detach(local_addr); });
		}

		void fault_handler(Signal_context_capability handler) override { _env.rm().fault_handler(handler); }
		State state() override { return _env.rm().state(); }
		Dataspace_capability dataspace() override { return _env.rm().dataspace(); }

		/***************
		 ** Accessors **
		 ***************/
		const Ram_quota_guard& ram_guard() const { return _ram_guard; }
		const Cap_quota_guard& cap_guard() const { return _cap_guard; }
		Env&                   env()       const { return _env; }
};
